/*************************************************************************
 * multipole_placement_rules.h - Handles multipole placement rules.      *
 *                                                                       *
 * Copyright (C) 2014-2015 by Laboratory of Medicinal Chemistry of       *
 * Department of Chemistry, Lomonosov Moscow State University            *
 *                                                                       *
 * Programmed by Oleg. I. Titov                                          *
 *                                                                       *
 * This file is part of the Electrostatic Tools project.                 *
 * For more information, see <http://qsar.chem.msu.ru/elec_tools/>       *
 *                                                                       *
 * Based on work:                                                        *
 * O.I. Titov, D.A. Shulga, V.A. Palyulin, N.S. Zefirov                  *
 * Description of halogen bonding on the basis of multicenter multipole  *
 * expansion // Doklady Chemistry, 2013, vol. 450, p. 139                *
 *                                                                       *
 * This program is free software; you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation version 2 of the License.                *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 ************************************************************************/

#ifndef MULTIPOLES_PLACEMENT_RULES_HEADER
#define MULTIPOLES_PLACEMENT_RULES_HEADER

#include <iostream>
#include <string>
#include <cctype>
#include <stdexcept>

#include "multipoles.h"
#include "comment_skipper.h"
#include "mult_tools_defines.h"

class MultipoleFlags
{
public:
  MultipoleFlags() : value(0) { }

  MultipoleFlags(const std::string s) : value(0)
  {
    if (!s.empty())
    {
      std::istringstream strm(s);
      readMe(strm);
    }
  }

  MultipoleFlags(unsigned int v) : value(v) { }

  MultipoleFlags(const MultipoleFlags& other) : value(other.value) { }

  bool test(MULTIPOLES v) const { return value & v; }

  void clear() { value = NONE; }

  void set(MULTIPOLES v) { value |= v; }

  void unset(MULTIPOLES v) { value &= ~v; }

  std::string readMe(std::istream& strm, std::string header = "")
  {
    char c;
    while (strm.good())
    {
      strm >> c;
      strm >> std::noskipws;
      switch (c)
      {
        case 'm' :
          set(MONOPOLE);
          break;
        case 'd' :
          set(DIPOLE);
          break;
        case 'q' :
          set(QUADRUPOLE);
          break;
        case '*' :
          set(SYMMETRIZED);
          break;
        case 'n' :
          value = NONE;
          break;
        case ' ' :
        case '\t' :
        case '\n' :
        case '\r' :
          strm >> std::skipws;
          return std::string();
        default :
          throw std::domain_error("Reading placement rules: Unknown multipole: " + c);
      }
    }
    return std::string();
  }

  void printMe(std::ostream& strm) const
  {
    if (test(MONOPOLE)) strm << "m";
    if (test(DIPOLE)) strm << "d";
    if (test(QUADRUPOLE)) strm << "q";
    if (test(SYMMETRIZED)) strm << "*";
  }

private:
  unsigned int value;
};

DECLARE_IO_STREAM_OPERATORS(MultipoleFlags)

class MultipolePlacementRule
{
public:
  MultipolePlacementRule(const std::string& s, MultipoleFlags r) : rule(r), SMARTS(s) { }
  
  MultipoleFlags rule;
  std::string SMARTS;
};

class MultipolePlacementRules
{
public:

  std::string readMe(std::istream& strm, std::string header = "")
  {
    if (header.empty()) strm >> header;
    if (header != "Placement-rules:") return header;
    clear();
    while (strm.good())
    {
      std::string type;
      strm >> type;
      if ((type == "atom:") || (type == "group:"))
      {
        std::string pattern;
        strm >> pattern;
        if (pattern.at(0) != '\"' || pattern.at(pattern.length() - 1) != '\"') throw std::domain_error("Reading placement rules: Bad atom SMARTS: " + pattern);
        MultipoleFlags rule;
        strm >> rule;
        if (type == "atom:") atomRules.push_back(MultipolePlacementRule(pattern.substr(1, pattern.length() - 2), rule));
        else groupRules.push_back(MultipolePlacementRule(pattern.substr(1, pattern.length() - 2), rule));
      } else return type;
    }
    return std::string();
  }

  void clear()
  {
    atomRules.clear();
    groupRules.clear();
  }
  
  int numAtomRules() const { return atomRules.size(); }

  int numGroupRules() const { return groupRules.size(); }

  const MultipolePlacementRule& atomRule(int i) const { return atomRules[i]; }

  const MultipolePlacementRule& groupRule(int i) const { return groupRules[i]; }

  void printMe(std::ostream& strm) const
  {
    strm << "Placement-rules:" << std::endl;
    for (int i = 0; i < atomRules.size(); ++i)
      strm << "atom: \"" << atomRules[i].SMARTS << "\" " << atomRules[i].rule << std::endl;
    for (int i = 0; i < groupRules.size(); ++i)
      strm << "atom: \"" << groupRules[i].SMARTS << "\" " << groupRules[i].rule << std::endl;
  }

private:
  std::vector< MultipolePlacementRule > atomRules;
  std::vector< MultipolePlacementRule > groupRules;
};

DECLARE_IO_STREAM_OPERATORS(MultipolePlacementRules)

#endif // MULTIPOLES_PLACEMENT_RULES_HEADER

